import sys
import os

from pathlib import Path

def add_path(new_path: str) -> None:
    """Add a new path to the system path if it is not already present.

    This function checks if the specified path is already in the system path. 
    If it is not, the function appends the new path to the system path, allowing for module imports from that location.

    Args:
        new_path (str): The path to be added to the system path.

    Returns:
        None
    """
    if new_path not in sys.path:
        sys.path.append(new_path)

def windows_to_wsl_path(windows_path):
    path = Path(windows_path).resolve()
    return f"/mnt/{path.drive[0].lower()}{str(path)[2:].replace('\\', '/')}"

def list_files_from_diretory(dir: str, endswith: str | None = None) -> list:
    """List files in a specified directory.

    This function retrieves all files from the given directory that match the specified suffix. 
    If no suffix is provided, it defaults to listing all files ending with '.json'.
    
    Args:
        dir (str): The path to the directory from which to list files.
        endswith (str | None): The suffix to filter files by. If None, all files are listed.

    Returns:
        list: A list of filenames that match the specified suffix.
    """
    return [f for f in os.listdir(dir) if f.endswith(endswith)]


class MigrationError(Exception):
    """
    Exceção personalizada para erros durante o processo de migração.
    """
    def __init__(self, message, details=None):
        super().__init__(message)
        self.details = details

    def __str__(self):
        if self.details:
            return f"{self.args[0]} | Detalhes: {self.details}"
        return self.args[0]



import os
import subprocess
import textwrap
import logging

from datetime import date 
from sqlalchemy import text, inspect, insert, MetaData, types, select
from sqlalchemy.orm import Session
from rich.progress import Progress
from utilities_pufm.prints.rich import custom_print
from utilities_pufm.database.db_connection.pg_connection import PgConnection
from sqlalchemy.engine import RowMapping
from utilities_pufm.files import files
from typing import List, Any, Dict, Union, Tuple

def execute_sql_query(sql_query, session: Session, params: dict = None, is_text: bool = False):
    
    try:
        if params:
            result = session.execute(sql_query if is_text else text(sql_query), params)
        else:
            result = session.execute(sql_query if is_text else text(sql_query))
        
        #TODO: Dve ser somente um tipo. Para não ter que fazer a conversão aqui para string.
        if is_text:
            sql_query = get_str_from_sqlalchemy_stmt(sql_query)
            # print(f"[EXECUTION-QUERY] Query:\n{sql_query}\n")
        
        if sql_query.strip().upper().startswith("SELECT"):
            results = result.fetchall()
            column_names = result.keys()
            
            return results, column_names
        elif sql_query.strip().upper().startswith("INSERT"):
            session.commit()    
            # custom_print(f"*[EXECUTION-QUERY]* Executando query:\n{sql_query}")
            
            insert_id = result.scalar()
            # custom_print(f"*[EXECUTION-QUERY]* ultimo ID:\n{insert_id}", colors=["red"])
            return insert_id, True
        
        elif any(sql_query.strip().upper().startswith(cmd) for cmd in ["UPDATE", "DELETE", "DROP", "REVOKE", "GRANT"]):
            session.commit()  # Commit necessário após alterações
            return result.rowcount, True
        
        elif any(sql_query.strip().upper().startswith(cmd) for cmd in ["SET", "ALTER"]):
            session.commit()
            return None, True
        
        elif sql_query.strip().upper().startswith("SHOW"):
            results = result.fetchall()
            return results, True
        
        return result.fetchall(), True
    
    except Exception as e:
        print(f"[EXECUTION] Erro na execução do script SQLAlchemy: {e}\n Query: {sql_query}")
        session.rollback()  # Faz o rollback em caso de erro
        files.save_file(f"./data/{date.today()}-ERRORS.txt", f"[ERROR]\n[EXECUTION] Erro na execução do script SQLAlchemy: {e}\n[SQL]: {sql_query}\n\n", how = "a")
        return f"Erro: {str(e)}", None
    

########################## Checking ##########################
def is_table_exists(session, table_name) -> bool:
    schema, table = table_name.split(".")
        
    query = textwrap.dedent("""
        SELECT COUNT(*)
        FROM information_schema.tables
        WHERE table_schema = :schema
        AND table_name = :table;
    """)

    result, _ = execute_sql_query(sql_query=query, session=session, params={"schema": schema, "table": escape_table_name(table)})
    
    return result[0][0] > 0

def is_line_existes(session: Session, referenced_table, referenced_column, value):
    schema, table = referenced_table.split(".")
    
    query = textwrap.dedent(f"""
        SELECT 1
        FROM {f'{schema}.' if schema else ''}{escape_table_name(table)}
        WHERE {referenced_column} = :value
        LIMIT 1;
    """)

    result = session.execute(query, {"value": value}).fetchone()

    return result is not None

def is_auto_increment(session: Session, table_name: str, column_name: str) -> bool:
    query = textwrap.dedent("""
    SELECT column_name, column_default
    FROM information_schema.columns
    WHERE table_schema = :schema
        AND table_name = :table_name
        AND column_name = :column_name;
    """)
    
    schema, table = table_name.split(".")
    result = execute_sql_query(sql_query=query, session=session, params={
        "schema": schema,
        "table_name": f"{schema}.{escape_table_name(table)}",
        "column_name": column_name}
    )
    
    return result is not None and result[1] is not None and 'nextval' in result[1]

def escape_table_name(table_name: str) -> str:
    if '@' in table_name:
        return f'"{table_name}"'
    return table_name

########################## Get Information ##########################
def get_all_rows(session, table_name):
    """
        Retrieve all rows from a specified table in the database.

        Args:
            session (Session): The SQLAlchemy session used to connect to the database.
            table_name (str): The name of the table from which to fetch all rows.

        Returns:
            tuple: A tuple containing the results as a list of rows and the column names.
    """
    schema, table = table_name.split(".")
    sql_query = textwrap.dedent(f"""
        SELECT *
        FROM {schema}.{escape_table_name(table)};
    """)
    return execute_sql_query(sql_query=sql_query, session=session)

def take_referenced_rows(session: Session, referenced_table: str, referenced_column: str, value: Union[str, List[str]]):
    """
    Retrieve all rows from a specified table in the database that matches the specified value(s)
    in the referenced column.

    Args:
        session (Session): The SQLAlchemy session used to connect to the database.
        referenced_table (str): The name of the table that contains the referenced column.
        referenced_column (str): The name of the column that contains the reference value.
        value (str or list): The value or list of values to filter the rows by.

    Returns:
        tuple: A tuple containing the results as a list of rows and the column names.
    """
    schema, table = referenced_table.split(".")
    query = textwrap.dedent(f"""
        SELECT *
        FROM {schema}.{escape_table_name(table)}
        WHERE {referenced_column} = :value;
    """) if not isinstance(value, list) else textwrap.dedent(f"""
        SELECT *
        FROM {schema}.{escape_table_name(table)}
        WHERE {referenced_column} = ANY(:value);
    """)
    
    return execute_sql_query(sql_query=query, session=session, params={'value': value})

def count_foreign_references(session, table_name) -> List:
    """
    Conta quantas vezes uma tabela é referenciada como chave estrangeira em outras tabelas.
    
    :param session: Sessão do SQLAlchemy.
    :param schema: Esquema da tabela.
    :param table: Nome da tabela.
    :return: Lista de dicionários contendo as tabelas que referenciam e a contagem de referências.
    """
    
    schema, table = table_name.split('.')
    query = textwrap.dedent("""
        SELECT 
            tc.table_schema, 
            tc.table_name, 
            COUNT(tc.constraint_name) AS reference_count
        FROM information_schema.table_constraints AS tc
        JOIN information_schema.key_column_usage AS kcu
            ON tc.constraint_name = kcu.constraint_name
            AND tc.table_schema = kcu.table_schema
        WHERE tc.constraint_type = 'FOREIGN KEY'
            AND kcu.table_schema = :schema
            AND kcu.table_name = :table
        GROUP BY tc.table_schema, tc.table_name
        ORDER BY reference_count DESC;
    """)
    
    result = session.execute(text(query), {'schema': schema, 'table': table}).fetchall()
    
    return [{'schema': row.table_schema, 'table': row.table_name, 'count': row.reference_count} for row in result]

def index_exists(session: Session, index_complete_name: str):
    schema, index_name = index_complete_name.split(".")
    result = session.execute(text(f"""
        SELECT indexname 
        FROM pg_indexes 
        WHERE schemaname = '{schema}' 
        AND indexname = '{index_name}';
    """))
    return result.scalar() is not None  # Retorna True se o índice existir

def drop_index(session: Session, index_complete_name: str):
    if index_exists(session=session, index_complete_name=index_complete_name):    
        session.execute(text(f"DROP INDEX {index_complete_name};"))
        session.commit()

def create_index(session: Session, index_complete_name: str, target_table: str, colums_short_name_list: List[str], where_clouse: str):
    
    _, index_name = index_complete_name.split(".")
    if not index_exists(session=session, index_complete_name=index_complete_name):
        session.execute(text(f"""
            CREATE UNIQUE INDEX {index_name} 
            ON {target_table}
            USING btree ({", ".join(column_short_name for column_short_name in colums_short_name_list)})
            {where_clouse};
        """))
        session.commit()

def get_all_fks(session, table_name):
    """
        Retrieve all foreign keys from a specified table in the database.

        Args:
            session (Session): The SQLAlchemy session used to connect to the database.
            table_name (str): The name of the table from which to fetch all foreign keys.

        Returns:
            list: A list of foreign key information, including the constraint name, constraint definition, source column, foreign key column, and foreign table.
    """
    schema, table = table_name.split(".")

    query_oid = textwrap.dedent("""
        SELECT oid
        FROM pg_class
        WHERE relname = :table
        AND relnamespace = (SELECT oid FROM pg_namespace WHERE nspname = :schema)
    """)

    result_oid, _ = execute_sql_query(sql_query=query_oid, session=session, params={"table": escape_table_name(table), "schema": schema})

    if not result_oid:
        print(f"[QUERY EXEC] Tabela {table_name} não encontrada.")
        return False

    return get_all_fks_for_table_oid(session, result_oid[0][0]) 

def get_all_fks_for_table_oid(session, table_oid):
    query_fks = textwrap.dedent("""
        SELECT
            conname AS constraint_name,
            pg_catalog.pg_get_constraintdef(r.oid, true) AS constraint_definition,
            att1.attname AS source_column,
            att2.attname AS foreign_key_column,
            nsp2.nspname || '.' || cl2.relname AS foreign_table
        FROM
            pg_catalog.pg_constraint r
        INNER JOIN
            pg_catalog.pg_class cl ON cl.oid = r.conrelid
        INNER JOIN
            pg_catalog.pg_attribute att1
            ON att1.attnum = ANY(r.conkey) AND att1.attrelid = r.conrelid
        INNER JOIN
            pg_catalog.pg_attribute att2
            ON att2.attnum = ANY(r.confkey) AND att2.attrelid = r.confrelid
        INNER JOIN
            pg_catalog.pg_class cl2 ON cl2.oid = r.confrelid
        INNER JOIN
            pg_catalog.pg_namespace nsp2 ON cl2.relnamespace = nsp2.oid
        WHERE
            r.conrelid = :oid AND r.contype = 'f'
        ORDER BY 1;
    """)
    
    return execute_sql_query(sql_query=query_fks, session=session, params={"oid": table_oid})

def get_primary_keys(session: Session, table_name):
    schema, table = table_name.split(".")
    
    inspector = inspect(session.get_bind())
    constraint_pk = inspector.get_pk_constraint(table, schema=schema)
    
    return constraint_pk["constrained_columns"] if constraint_pk["constrained_columns"] else None

def update_null_inline(
    session: Session,
    table_name: str,
    attribute_name: Union[str, List[str]],
    attribute_value: Union[str, int, List[Union[str, int]]],
    data_dict: Dict
) -> Tuple[bool, str]:

    """
    Execute an update statement with the given set and where clauses.

    Args:
        session (Session): The SQLAlchemy session used to connect to the database.
        table_name (str): The name of the table to update.
        attribute_name (Union[str, List[str]]): The name of the column(s) to update.
        attribute_value (Union[str, int, List[Union[str, int]]]): The value(s) to update the column(s) with.
        data_dict (Dict): A dictionary containing the set clauses.

    Returns:
        Tuple[bool, str]: A tuple containing a boolean indicating if the operation was successful and a string with the query executed.
    """
    
    set_clauses = []
    params = {}

    if isinstance(attribute_name, str):
        attribute_name = [attribute_name]
    if not isinstance(attribute_value, list):
        attribute_value = [attribute_value]

    if len(attribute_name) != len(attribute_value):
        raise ValueError("The number of columns in attribute_name does not match the number of values in attribute_value.")

    for column, value in data_dict.items():
        if value is not None:
            set_clauses.append(f"{column} = COALESCE({column}, :{column})")
            params[column] = value

    where_clauses = []
    for i, (col, val) in enumerate(zip(attribute_name, attribute_value)):
        placeholder = f"attr_{i}"
        where_clauses.append(f"{col} = :{placeholder}")
        params[placeholder] = val

    if set_clauses:
        query = textwrap.dedent(f"""
            UPDATE {table_name}
            SET {', '.join(set_clauses)}
            WHERE {' AND '.join(where_clauses)}
        """)

        return execute_sql_query(sql_query=query, session=session, params=params)
    return -1, False

def get_real_query(check_query, params):
    sql_real = check_query
    for key, value in params.items():
        sql_real = sql_real.replace(f":{key}", f"'{value}'")

    return sql_real

def is_foreign_key_exists(session: Session, table: str, foreign_key_name: str):
    schema, table_name = table.split(".")
    check_query = textwrap.dedent("""
        SELECT 1 
        FROM information_schema.table_constraints 
        WHERE table_schema = :schema
            AND table_name = :table 
            AND constraint_name = :foreign_key_name 
            AND constraint_type = 'FOREIGN KEY';
    """)

    params = {"schema": schema, "table": table_name, "foreign_key_name": foreign_key_name}
    result = session.execute(text(check_query), params).scalar()
    
    # custom_print(f"*[TESTE-is_foreign_key_exists]*: Query: \n{get_real_query(check_query, params)}'\n\n'{bool(result)}'", colors=["red"])
    
    return bool(result)

def drop_foreign_constraint(session: Session, table: str, foreign_key_name: str):
    
    sql_query = textwrap.dedent(f"""
        ALTER TABLE {table}
        DROP CONSTRAINT IF EXISTS {foreign_key_name};
    """)

    return execute_sql_query(sql_query=sql_query, session=session)

def get_max_column_value(session, table_name: str, column_name: str):
    
    sql_query = textwrap.dedent(f"""
        SELECT MAX({column_name}) AS max_value
        FROM {table_name};
    """)
    
    max_val = execute_sql_query(sql_query=sql_query, session=session)
    
    # print(max_val[0][0])
    return max_val[0][0][0] if max_val else None

def get_column_values(session, table_name: str, column_name: str):
    query = select(text(column_name)).select_from(text(table_name))
    result = session.execute(query).scalars().all()
    
    return result if result else []
    
def recreate_foreign_constraint(session, table_origem: str, foreign_key_name: str, column_origem: str, table_destino: str, column_destino: str):
    sql_query = textwrap.dedent(f"""
        ALTER TABLE {table_origem}
        ADD CONSTRAINT {foreign_key_name}
        FOREIGN KEY ({column_origem}) 
        REFERENCES {table_destino}({column_destino});
    """)

    return execute_sql_query(sql_query=sql_query, session=session)

def get_columns_names_and_types(session, table_name):
    
    schema, table = table_name.split(".")
    
    inspector = inspect(session.get_bind())
    columns = inspector.get_columns(table, schema=schema)

    return [(col['name'], col['type']) for col in columns]

def get_str_from_sqlalchemy_stmt(stmt):
    return str(stmt.compile())

########################## Set information ##########################
def insert_row(session: Session, columns: List[str], values: List[Any], table_name: str) -> None:
    schema, table = table_name.split(".")
    
    if len(columns) != len(values):
        raise ValueError("The number of values on columns and values must be the same.")

    engine = session.get_bind()
    if not hasattr(engine, '_metadata_cache'):
        engine._metadata_cache = {}

    table_escaped = escape_table_name(table)
    
    metadata_key = f"{schema}.{table_escaped}" if schema else table_escaped
    if metadata_key not in engine._metadata_cache:
        metadata = MetaData()
        metadata.reflect(schema=schema, bind=engine, only=[table_escaped])
        engine._metadata_cache[metadata_key] = metadata
    else:
        metadata = engine._metadata_cache[metadata_key]

    qualified_name = f"{schema}.{table_escaped}" if schema else table_escaped
    table_obj = metadata.tables.get(qualified_name)
    if table_obj is not None:
        try:
            data = dict(zip(columns, values))
            stmt = insert(table_obj).values(data).returning(table_obj.primary_key.columns)
            result = execute_sql_query(sql_query=stmt, session=session, is_text=True)
            return result, get_real_query(get_str_from_sqlalchemy_stmt(stmt), data)
        except Exception as e:
            logging.error(f"Erro ao inserir na tabela {table_name}: {str(e)}")
            raise
    else:    
        raise ValueError(f"Table '{qualified_name}' not found")

########################## Backups e Restaurations ##########################
def backup_all_individual_tables(
    conn,
    backup_dir: str,
    verbose: bool = False,
    progress_bar: Progress = None,
    return_verbose: bool = False
) -> bool:

    os.makedirs(backup_dir, exist_ok=True)
    session_id, session = conn.get_new_session()
    verbose_list = []
    status_list = []
    
    tables, _ = execute_sql_query(
        sql_query=textwrap.dedent(
            """
            SELECT schemaname, tablename
            FROM pg_tables
            WHERE schemaname NOT IN ('pg_catalog', 'information_schema');
            """
        ),
        session=session
    )
    progress_bar.start()
    backup_task = progress_bar.add_task(description="[bold yellow]Backup[/bold yellow] em progresso...", total=len(tables))
    for schema_name, table_name in tables:
        
        bkp_result = backup_table(
            conn=conn,
            schema_name=schema_name,
            table_name=table_name,
            backup_dir=backup_dir,
            verbose=verbose,
            return_verbose=return_verbose
        )
        if return_verbose:
            status_bkp, verbose_list_table, _ = bkp_result
        else:
            status_bkp, _ = bkp_result
        
        verbose_list.extend(verbose_list_table)
        status_list.append(status_bkp)
        
        progress_bar.update(backup_task, advance=1)
    progress_bar.stop()
        
    if any(not status for status in status_list):
        verbose_str = "Pelo menos um 'backup' falhou."
    else:
        verbose_str = "Backup de todas as tabelas de todos os esquemas concluído com sucesso."
    if verbose:
        print(verbose_str)
    if return_verbose:
        verbose_list.append(verbose_str)
    
    conn.close_session(session_id)
    if return_verbose:
        return True, verbose_list
    return True

def backup_table(conn, db: str, schema_name: str, table_name: str, backup_dir: str, verbose: bool = False, return_verbose: bool = False):
    verbose_list = []
    
    backup_file_name = f"{db}_{schema_name}_{table_name}_backup.sql"
    backup_file_path = os.path.join(backup_dir, backup_file_name)

    engine = conn.get_engine()
    
    verbose_str = f"Iniciando backup da tabela {schema_name}.{table_name}..."
    if verbose:
        print(verbose_str)
    if return_verbose:
        verbose_list.append(verbose_str)

    connection_info = engine.url
    command = (
        f"pg_dump -h {connection_info.host} -U {connection_info.username} -d {connection_info.database} "
        f"-t {schema_name}.{table_name} -f {backup_file_path} --clean --if-exists --column-inserts --disable-triggers"
    )

    result = subprocess.run(command, shell=True)

    if result.returncode != 0:
        verbose_str = f"Erro ao fazer backup da tabela {schema_name}.{table_name}."
        print(verbose_str)
        if return_verbose:
            verbose_list.append(verbose_str)
            return False, verbose_list, backup_file_path
        else:
            return False, backup_file_path

    verbose_str = f"Backup da tabela {schema_name}.{table_name} concluído com sucesso."
    if verbose:
        print(verbose_str)
    if return_verbose:
        verbose_list.append(verbose_str)
        return True, verbose_list, backup_file_path
    else:
        return True, backup_file_path

def backup_all_database(
    backup_dir: str,
    db_name: str,
    conn = None,
    verbose: bool = False,
    return_verbose: bool = False,
    pg_path: str = None,
    params: dict = None,
    include_roles: bool = True
) -> bool:

    # print(f"[PARAMETERS]:\n{params}")
    
    if conn is None and params is None:
        raise ValueError("Nenhuma conexão ou parâmetros de conexão fornecidos.")
    
    if conn is not None:
        engine = conn.get_engine()    
        connection_info = engine.url
        host = connection_info.host
        username = connection_info.username
        os.environ['PGPASSWORD'] = connection_info.password
        database = connection_info.database
    elif params is not None:
        host = params['host']
        username = params['username']
        database = params['database']
        os.environ['PGPASSWORD'] = params['password']
        
    os.makedirs(backup_dir, exist_ok=True)

    backup_file_name = f"{db_name}_backup-all.dump"
    backup_file_path = os.path.join(backup_dir, backup_file_name)
    
    verbose_list = []
    verbose_str = f"Iniciando backup completo do banco de dados {db_name}..."
    if verbose:
        print(verbose_str)
    if return_verbose:
        verbose_list.append(verbose_str)
    
    pg_bin_path = ''
    if pg_path:
        pg_bin_path = f'"{pg_path}"'
    
    dump_command = (
        f"{pg_bin_path}pg_dump -h {host} -U {username} -d {database} "
        f"-f {backup_file_path} -Fc --clean --if-exists --column-inserts --create"
    )

    dump_result = subprocess.run(dump_command, shell=True)

    if include_roles:
        role_backup_file_name = "roles_backup-all.dump"    
        role_backup_file_path = os.path.join(backup_dir, role_backup_file_name)
        
        roles_command = (
            f"{pg_bin_path}pg_dumpall -h {host} -U {username} --roles-only "
            f"-f {role_backup_file_path}"
        )

        roles_result = subprocess.run(roles_command, shell=True)
    
    if dump_result.returncode != 0:
        verbose_str = f"Erro ao fazer backup completo do banco de dados {db_name}."
        if verbose:
            print(verbose_str)
        if return_verbose:
            verbose_list.append(verbose_str)
            
        return False, verbose_list
    elif include_roles and roles_result.returncode != 0:
        verbose_str = f"Erro ao fazer backup dos papéis do banco de dados {db_name}."
        if verbose:
            print(verbose_str)
        if return_verbose:
            verbose_list.append(verbose_str)
            
        return False, verbose_list

    if verbose:
        verbose_str = f"Backup completo do banco de dados {db_name} concluído com sucesso. Backup salvo em {backup_file_path}."
        if include_roles:
            verbose_str += f" Backup dos papéis também concluído. Backup salvo em {role_backup_file_path}."
        
        if verbose:
            print(verbose_str)
        if return_verbose:
            verbose_list.append(verbose_str)

    if return_verbose:
        return True, verbose_list
    return True

def restore_all_database(conn, backup_dir: str, bd_name: str, new_terminal: bool = False, verbose: bool = True, return_verbose: bool = False, pg_path: str = None, include_roles: bool = True) -> bool:
    
    file_restore_path = os.path.join(backup_dir, bd_name + "_backup-all.dump")
    
    engine =  conn.get_engine()
    connection_info = engine.url
    if return_verbose:
        verbose_list = []
    
    verbose_str = f"[RESTORE]: Initializing restore process from {file_restore_path}... "
    if verbose:
        print(verbose_str)
    if return_verbose:
        verbose_list.append(verbose_str)
    
    conn.close_all_sessions()
    
    verbose_str = "[RESTORE]: Trying to restore database..."
    if verbose:
        print(verbose_str)
    if return_verbose:
        verbose_list.append(verbose_str)
    
    pg_bin_path = ''
    if pg_path:
        pg_bin_path = f'"{pg_path}"'
        
    dropdb_command = f"{pg_bin_path}dropdb -h {connection_info.host} -U {connection_info.username} --force {connection_info.database}"
    result = subprocess.run(dropdb_command, shell=True, text=True, capture_output=True)
    
    if result.returncode != 0:
        verbose_str = f"[RESTORE]: Warning during drop database: {result.stderr}"
        if verbose:
            print(verbose_str)
        if return_verbose:
            verbose_list.append(verbose_str)
    else:
        verbose_str = "[RESTORE]: Database dropped successfully"
        if verbose:
            print(verbose_str)
        if return_verbose:
            verbose_list.append(verbose_str)
    
    if include_roles:
        role_backup_file_name = "roles_backup-all.dump"    
        role_backup_file_path = os.path.join(backup_dir, role_backup_file_name)
        
        roles_command = (
            f"{pg_bin_path}psql -h {connection_info.host} -U {connection_info.username} -f {role_backup_file_path}"
        )

        roles_result = subprocess.run(roles_command, shell=True)
    
        if roles_result.returncode != 0:
            verbose_str = f"[RESTORE-restore]: Warning during roles restoring: {roles_result.stderr}"
            if verbose:
                print(verbose_str)
            if return_verbose:
                verbose_list.append(verbose_str)
        else:
            verbose_str = "[RESTORE-restore]: Roles restored successfully!"
            if verbose:
                print(verbose_str)
            if return_verbose:
                verbose_list.append(verbose_str)
    
    createdb_command = f"{pg_bin_path}createdb -h {connection_info.host} -U {connection_info.username} {connection_info.database}"
    result = subprocess.run(createdb_command, shell=True, text=True, capture_output=True)
    
    if result.returncode != 0:
        verbose_str = f"[RESTORE]: Warning during create database: {result.stderr}"
        if verbose:
            print(verbose_str)
        if return_verbose:
            verbose_list.append(verbose_str)
    else:
        verbose_str = "[RESTORE]: Database created successfully"
        if verbose:
            print(verbose_str)
        if return_verbose:
            verbose_list.append(verbose_str)
    
    command = f"{pg_bin_path}pg_restore --disable-triggers -v --clean --if-exists --create --jobs=10 -U postgres -d postgres {file_restore_path}"

    if new_terminal:
        command = f'start cmd /K "{command}"'

    try:
        result = subprocess.run(command, shell=True, text=True, capture_output=True)
        
        success = result.returncode == 0
        status_msg = "[RESTORE]: Restore to database was completed successfully." if success else "[RESTORE]: Restore failed."
        
        if verbose:
            print(status_msg)
        if return_verbose:
            verbose_list.append(status_msg)
            return success, verbose_list
            
        return success
        
    except Exception as e:
        error_msg = f"[RESTORE]: An error occurred during restore: {str(e)}"
        if verbose:
            print(error_msg)
        if return_verbose:
            verbose_list.append(error_msg)
            return False, verbose_list
        return False
    
    finally:
        #TODO: Verificar qual é a sessião que deve continuar aberta
        conn.get_new_session()
        verbose_str = "[RESTORE]: Session restarted. Restore terminated succefully."
        if verbose:
            print(verbose_str)
        if return_verbose:
            verbose_list.append(verbose_str)

def restore_table(conn: PgConnection, backup_file: str, table: str, pg_path: str = None, verbose: bool = True, return_verbose: bool = False):
    
    inspector = inspect(conn.get_engine())
    
    if table in inspector.get_table_names():
        session_id, session = conn.get_new_session()
        session.execute(text(f"DELETE FROM {table};"))
        session.commit()
        conn.close_session(session_id=session_id)
        print(f"All registers from {table} were deleted.")
    else:
        print(f"Table {table} does not exist in the database {conn.get_engine().url}.")
    
    pg_bin_path = ''
    if pg_path:
        pg_bin_path = f'"{pg_path}"'
    
    #* restaurar o estado da tabela anterior
    command = f"{pg_bin_path}pg_restore --disable-triggers -v --clean --if-exists --create --jobs=10 -U postgres -d postgres -t {table} {backup_file}"
    
    if return_verbose:
        verbose_list = []
    
    try:
        result = subprocess.run(command, shell=True, text=True, capture_output=True)

        success = result.returncode == 0
        
        status_msg = "[RESTORE]: Restore to database was completed successfully." if success else "[RESTORE]: Restore failed."
    
        if verbose:
            print(status_msg)
        if return_verbose:
            verbose_list.append(status_msg)
            return success, verbose_list
            
        return success
        
    except Exception as e:
        error_msg = f"[RESTORE]: An error occurred during restore: {str(e)}"
        if verbose:
            print(error_msg)
        if return_verbose:
            verbose_list.append(error_msg)
            return False, verbose_list
        return False
    
    finally:
        
        verbose_str = f"[RESTORE]: Table {table} restored successfully."
        if verbose: 
            print(verbose_str)
        if return_verbose:
            verbose_list.append(verbose_str)
            return success, verbose_list
        return success
    
    


#* Imports
import os
#* Imports dos módulos
from utilities_pufm.database.enums import SGBDType
from utilities_pufm.database.db_connection.pg_connection import PgConnection
from utilities_pufm.database.db_connection.mysql_connection import MySQLConnection

def get_connection(name: str, **kargs):
     sgbd_type = os.getenv(f"{name}_TYPE")
     if sgbd_type == SGBDType.TYPE_MYSQL.value:
          return MySQLConnection(name=name, **kargs)
     elif sgbd_type == SGBDType.TYPE_POSTGRES.value:
          return PgConnection(name=name, **kargs)
     print("[CONNECTION]: Tipo de conecção desconhecida.")
     return None


from rich.tree import Tree
from rich.console import Console
import os

def build_tree(path, tree):
    """
    Recursively builds a rich Tree object for the directory structure.
    
    :param path: Current directory path.
    :param tree: Rich Tree object to build upon.
    """
    for entry in os.scandir(path):
        if entry.is_dir():
            # Add directory to the tree
            branch = tree.add(f"[bold blue]{entry.name}[/]")
            # Recurse into subdirectory
            build_tree(entry.path, branch)
        else:
            # Add file to the tree
            tree.add(f"[green]{entry.name}[/]")

def print_directory_tree(root_path):
    """
    Prints the directory tree using rich.

    :param root_path: Path to the directory to visualize.
    """
    # Create the root of the tree
    tree = Tree(f"[bold magenta]{os.path.basename(root_path)}[/]")
    build_tree(root_path, tree)
    # Print the tree
    console = Console()
    console.print(tree)
    

""""
precisa verificar

"""


def create_json_directory_tree(base_path, tree_structure):
    """
    Recursively creates a directory tree.

    :param base_path: The root path where the tree will be created.
    :param tree_structure: A nested dictionary representing the directory structure.
    """
    for directory, sub_dirs in tree_structure.items():
        # Define the path for the current directory
        dir_path = os.path.join(base_path, directory)
        
        # Create the current directory
        os.makedirs(dir_path, exist_ok=True)
        
        # If there are subdirectories, create them recursively
        if isinstance(sub_dirs, dict):
            create_directory_tree(dir_path, sub_dirs)

    # Define the directory tree structure as a nested dictionary
    directory_tree = {
        "root_folder": {
            "sub_folder1": {
                "sub_sub_folder1": {},
                "sub_sub_folder2": {},
            },
            "sub_folder2": {
                "sub_sub_folder3": {},
            },
            "sub_folder3": {}
        }
    }

    # Base path where the directory tree will be created
    base_path = "my_project"

    # Generate the directory tree
    create_directory_tree(base_path, directory_tree)

    print(f"Directory tree created at: {os.path.abspath(base_path)}")



import json

import numpy as np
import pandas as pd

from datetime import datetime, date
from decimal import Decimal
from uuid import UUID
from typing import Dict

def load_json(json_path: str) -> object:
    """Load data from a JSON file.

    This function reads a JSON file from the specified path and returns the data as a Python object. 
    It handles errors related to file not found and invalid JSON format, providing informative messages.

    Args:
        json_path (str): The path to the JSON file to be loaded.

    Returns:
        Any: The data loaded from the JSON file.

    Raises:
        FileNotFoundError: If the specified file does not exist.
        json.JSONDecodeError: If the file does not contain valid JSON.
    """

    try:
        with open(json_path, 'r', encoding='utf-8') as file:
            data = json.load(file)
        return data
    except FileNotFoundError:
        print(f"Erro: O arquivo '{json_path}' não foi encontrado.")
        raise
    except json.JSONDecodeError as e:
        print(f"Erro: O arquivo '{json_path}' não contém um JSON válido.\n{e}")
        raise

def load_or_create_json(json_path: str, data: dict = {}) -> Dict:
    """Load data from a JSON file or create a new file if it doesn't exist.

    This function attempts to load data from an existing JSON file at the specified path. 
    If the file does not exist, it creates a new empty file and returns an empty dictionary.
    It handles errors related to file not found and invalid JSON format, providing informative messages.

    Args:
        json_path (str): The path to the JSON file to be loaded or created.

    Returns:
        Any: The data loaded from the JSON file or an empty dictionary if the file doesn't exist.

    Raises:
        FileNotFoundError: If the specified file does not exist.
        json.JSONDecodeError: If the file does not contain valid JSON.
    """

    try:    
        return load_json(json_path)
    except FileNotFoundError:
        save_json(json_path, data)
        return data   
    except json.JSONDecodeError as e:
        print(f"Erro: O arquivo '{json_path}' não contém um JSON válido.\n{e}")
        raise

def save_json(file_path, data):
    """Save data to a JSON file.

    This function writes the provided data to a specified file in JSON format. 
    The data is formatted with an indentation of four spaces for better readability.

    Args:
        file_path (str): The path to the file where the JSON data will be saved.
        data (Any): The data to be saved in JSON format.

    Returns:
        None
    """

    with open(file_path, "w") as file:
        json.dump(data, file, indent=4)

def make_obj_json_serializable(obj):

    def handle_dict(dict):
        return {key: make_obj_json_serializable(value) for key, value in dict.items()}
    
    # Handle None
    if obj is None:
        # print("[JSON WARNING]: Object is None.")
        return None
        
    # Handle basic types that are already serializable
    if isinstance(obj, (bool, int, float, str)):
        # print("[JSON WARNING]: Object is already serializable.")
        return obj
    
    # Handle numpy numeric types
    if isinstance(obj, (np.int_, np.intc, np.intp, np.int8, np.int16, np.int32, np.int64, np.uint8, np.uint16, np.uint32, np.uint64)):
        return int(obj)
    
    if isinstance(obj, (np.float16, np.float32, np.float64)):
        return float(obj)
    
    # Handle numpy arrays
    if isinstance(obj, np.ndarray):
        return obj.tolist()
    
    # Handle datetime objects
    if isinstance(obj, (datetime, date)):
        return obj.isoformat()
    
    # Handle Decimal
    if isinstance(obj, Decimal):
        return str(obj)
    
    # Handle UUID
    if isinstance(obj, UUID):
        return str(obj)
    
    # Handle pandas Series/DataFrame
    if isinstance(obj, (pd.Series, pd.DataFrame)):
        return handle_dict(obj.to_dict())
        
    # Handle dictionaries recursively
    if isinstance(obj, dict):
        return handle_dict(obj)
        
    # Handle iterables (lists, tuples, sets)
    if isinstance(obj, (list, tuple, set)):
        return [make_obj_json_serializable(item) for item in obj]
    
    # Handle any other objects by converting to string
    try:
        return str(obj)
    except Exception:
        return f"<non-serializable: {type(obj).__name__}>"
    


import os

def update_file(file_path: str, content: str) -> None:
    with open(file_path, "a", encoding="utf-8") as f:
        f.write(content)
        
def save_file(file_path: str, content: str, how: str = "w") -> None:
    
    directory = os.path.dirname(file_path)
    
    create_dir_if_not_exist(dir_path=directory)
        
    with open(file_path, how, encoding="utf-8") as f:
        f.write(content)
        
def create_dir_if_not_exist(dir_path: str) -> None:
    if dir_path and (not os.path.exists(dir_path)):
        os.makedirs(dir_path, exist_ok=True)


import sys
from io import StringIO
from collections import deque

class LogRedirector:
    def __init__(self, update_log_func):
        """
        Redireciona mensagens de print para o painel de log customizado.

        Args:
            update_log_func (callable): Função para atualizar o painel de log.
        """
        self.update_log_func = update_log_func
        self.original_stdout = sys.stdout
        self.buffer = StringIO()

    def write(self, message):
        # Remove novas linhas para evitar mensagens vazias
        if message.strip():
            self.update_log_func(message.strip())

    def flush(self):
        """Necessário para compatibilidade com sys.stdout."""
        pass

    def start(self):
        """Inicia o redirecionamento."""
        sys.stdout = self

    def stop(self):
        """Restaura o stdout original."""
        sys.stdout = self.original_stdout
        
class RichLogger():
    _instance = None
    def __init__(self):
        self.messages = deque(["[bold white]Log:[/bold white]"])
        self.size = 20

    def __new__(class_, *args, **kwargs):
        if not isinstance(class_._instance, class_):
            class_._instance = object.__new__(class_, *args, **kwargs)
        return class_._instance

    def write(self, message):
        self.messages.extend(message.splitlines())
        while len(self.messages) > self.size:
            self.messages.popleft()

    def flush(self):
        pass


import sys
import io
from contextlib import contextmanager

@contextmanager
def redirect_stdout(f):
    try:
        _stdout = sys.stdout
        sys.stdout = f
        yield
    finally:
        sys.stdout = _stdout


    f = io.StringIO()
    with redirect_stdout(f):
        do_something()
    out = f.getvalue()


    from io import StringIO 
    import sys

class Capturing(list):
    def __enter__(self):
        self._stdout = sys.stdout
        sys.stdout = self._stringio = StringIO()
        return self
    def __exit__(self, *args):
        self.extend(self._stringio.getvalue().splitlines())
        del self._stringio    # free up some memory
        sys.stdout = self._stdout


import pandas as pd
from ydata_profiling import ProfileReport

def save_and_report(df: pd.DataFrame, save_path: str) -> None:
    """Generates a data profiling report and saves it to a file.

    This function uses the `ydata_profiling` library to create a comprehensive
    profiling report for the provided DataFrame. The report is saved as a file
    at the specified path.

    Args:
        df (pd.DataFrame): The input DataFrame for which the profiling report is generated.
        save_path (str): The file path where the profiling report will be saved.
            The path should include the file name and extension (e.g., `.html`).

    Returns:
        None

    Raises:
        FileNotFoundError: If the specified path is invalid or cannot be accessed.
        ValueError: If the DataFrame is empty or invalid for profiling.

    Example:
        >>> import pandas as pd
        >>> from ydata_profiling import ProfileReport
        >>> df = pd.DataFrame({'A': [1, 2, 3], 'B': [4, 5, 6]})
        >>> save_and_report(df, "output_report.html")
    """
    profile = ProfileReport(df, title="report")
    profile.to_file(save_path)
    
def save_csv_file(df: pd.DataFrame, save_path: str) -> None:
    """Saves a DataFrame to a CSV file.

    This function writes the contents of the provided DataFrame to a CSV file
    at the specified path.

    Args:
        df (pd.DataFrame): The input DataFrame to be saved as a CSV file.
        save_path (str): The file path where the CSV file will be saved.
            The path should include the file name and `.csv` extension.

    Returns:
        None

    Raises:
        FileNotFoundError: If the specified path is invalid or cannot be accessed.
        PermissionError: If the program lacks the necessary permissions to write to the file.
        ValueError: If the DataFrame is empty or invalid for saving.

    Example:
        >>> import pandas as pd
        >>> df = pd.DataFrame({'A': [1, 2, 3], 'B': [4, 5, 6]})
        >>> save_csv_file(df, "output_data.csv")
    """
    df.to_csv(path_or_buf=save_path)


from enum import Enum

class Verbosetype(Enum):
    TYPE_DEFAULT="default"
    TYPE_RICH="rich"
    
class RichColors(Enum):
    RED = "red"
    GREEN = "green"
    YELLOW = "yellow"
    BLUE = "blue"
    MAGENTA = "magenta"
    CYAN = "cyan"
    WHITE = "white"
    BLACK = "black"


from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
import os

class MySQLConnection:
    def __init__(self, name):
        try:
            print(f"[MYSQL-CONECTION]: Iniciando processo de conexão com {name}...")

            # Obtenção dos parâmetros de conexão a partir de variáveis de ambiente
            self.connection_params = {
                "user": os.getenv(f"{name}_USER"),
                "password": os.getenv(f"{name}_PASSWORD"),
                "host": os.getenv(f"{name}_HOST"),
                "port": os.getenv(f"{name}_PORT"),
                "database": os.getenv(f"{name}_DB_NAME"),
            }

            # Verificar se todos os parâmetros estão presentes
            for key, value in self.connection_params.items():
                if value is None:
                    raise ValueError(f"[MYSQL-CONECTION]: Parâmetro {key} não encontrado nas variáveis de ambiente")

            # Construir a URL de conexão para o SQLAlchemy
            self.connection_url = (
                f"mysql+pymysql://{self.connection_params['user']}:{self.connection_params['password']}"
                f"@{self.connection_params['host']}:{self.connection_params['port']}/{self.connection_params['database']}"
            )
            print(f"[MYSQL-CONECTION]: Conectando usando URL: {self.connection_url}")

            # Criar engine do SQLAlchemy
            self.engine = create_engine(self.connection_url)
            self.Session = sessionmaker(bind=self.engine)

            # Testar conexão
            with self.engine.connect() as conn:
                print("[MYSQL-CONECTION]: Conexão estabelecida com sucesso.")

            self.session = None  # Sessão será criada sob demanda

        except Exception as e:
            print(f"[MYSQL-CONECTION]: Falha na conexão. Erro: {e}")
            raise

    def get_session(self):
        """Cria e retorna uma nova sessão."""
        if self.session is None:
            self.session = self.Session()
            print("[MYSQL-CONECTION]: Sessão criada.")
        return self.session

    def close_session(self):
        """Fecha a sessão, se existir."""
        if self.session:
            self.session.close()
            print("[MYSQL-CONECTION]: Sessão fechada.")
            self.session = None

    def close_engine(self):
        """Fecha o engine do SQLAlchemy."""
        if self.engine:
            self.engine.dispose()
            print("[MYSQL-CONECTION]: Engine fechado.")

    def close_all(self):
        """Fecha sessão e engine."""
        self.close_session()
        self.close_engine()



import itertools
import logging
import os

from io import StringIO
from rich.tree import Tree
from rich import print as rich_print
from utilities_pufm.utils.enums import RichColors
from utilities_pufm.prints.logs import RichLogger, LogRedirector
from datetime import datetime
from rich.logging import RichHandler
from rich import box, print_json
from rich.align import Align
from rich.console import Console
from rich.layout import Layout
from rich.panel import Panel
from rich.progress import BarColumn, Progress, SpinnerColumn, TextColumn, MofNCompleteColumn, TimeRemainingColumn, TimeElapsedColumn
from rich.syntax import Syntax
from rich.table import Table

def custom_print_json(data):
    print_json(data=data)    

def print_matriz_bonita(df, nome: str = ""):
    console_terminal = Console()
    table = Table(title=f"Matriz de Similaridade entre Colunas - {nome}", box=box.ROUNDED, show_lines=True)

    table.add_column("↓ columns_to_drop / columns_to_add →", style="bold magenta", no_wrap=True)
    for col in df.columns:
        table.add_column(col, style="cyan", justify="center")

    for idx, row in df.iterrows():
        row_values = [f"{val:.1f}%" for val in row]
        table.add_row(idx, *row_values)

    buffer = StringIO()
    console = Console(file=buffer, force_terminal=True, width=140, no_color=True)
    console.print(table)
    console_terminal.print(table)
    
    return buffer.getvalue()

def custom_print(msg: str, rich: bool = True, colors: list = ["yellow"]):
    """
    Prints customized messages with optional styling using the Rich library.

    Args:
        msg (str): The message to display. 
                   To style parts of the text, use the '*' character as a delimiter.
                   For example, '*styled text*' will be displayed with special formatting.
        rich (bool): Determines whether the Rich library should be used for styling the output.
        colors (list): A list of colors to alternate for styling text segments 
                       enclosed by '*'. Colors should be strings and correspond to 
                       the `RichColors` enum.

    Example:
        custom_print("Normal text, *styled text*", rich=True, colors=["red", "blue"])

    Note:
        When using the Rich feature, the provided color names must be valid 
        according to the `RichColors` enum.
    """
    if rich:
        colors = itertools.cycle([RichColors[color.upper()].value for color in colors]) 
        first = True
        rich_msg = ""
        current_color = next(colors)
        for char in msg:
            if char == "*" and first:
                rich_msg += f"[bold {current_color}]"
                first = False
            elif char == "*":
                rich_msg += f"[/bold {current_color}]"
                first = True
                current_color = next(colors)
            else:
                rich_msg += char
                        
        rich_print(rich_msg)
        
    else:
        print(msg)

def custom_progress_bar(console: Console = None) -> Progress:
    """
    Creates and returns a custom progress bar using the Rich library.

    The progress bar includes:
        - A text column displaying the task description.
        - A text column showing the task completion percentage.
        - A bar column representing the progress visually.
        - A column showing the number of completed tasks out of the total tasks.
        - Time elapsed since the task started.
        - Estimated time remaining to complete the task.

    Returns:
        Progress: An instance of Rich's Progress class configured with the specified columns.
    """

    return Progress(
            TextColumn("[progress.description]{task.description}"), 
            TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
            BarColumn(),
            MofNCompleteColumn(),
            TextColumn("•"),
            TimeElapsedColumn(),
            TextColumn("•"),
            TimeRemainingColumn(),
            console=console
        )

def print_directory_structure(directory="."):
    def build_tree(directory, tree):
        try:
            for entry in sorted(os.listdir(directory)):
                path = os.path.join(directory, entry)
                if os.path.isdir(path):
                    subtree = tree.add(f"[bold blue]📁 {entry}[/]")
                    build_tree(path, subtree)
                else:
                    tree.add(f"[green]📄 {entry}[/]")
        except PermissionError:
            tree.add("[red]⛔ Permission denied[/]")

    tree = Tree(f"[bold yellow]📂 {directory}[/]")
    build_tree(directory, tree)
    rich_print(tree)
    

class CustomRichLayout():
    FORMAT = '%(asctime)s'

    def __init__(self, name_layout, init_job_list, *args,**kwargs):
        self.name_layout = name_layout
        self.init_job_list = init_job_list
        self.log_message = []
        self.layout = None
        
        self.rl = RichLogger()
        self.lr = LogRedirector(update_log_func=self.rl.write)
        
        self.c = Console(file=self.rl, width=150)
        r = RichHandler(console=self.c)
        logging.basicConfig(
            level="NOTSET", format=self.FORMAT, datefmt="[%X]", handlers=[r]
        )
        self.logger = logging.getLogger("rich")
        
        self._define_layout()
        
    def _make_log_panel(self) -> Panel:
        """Some example content."""
        # log_message = Table.grid(padding=1)
        # log_message.add_column(style="green", justify="right")
        # log_message.add_column(no_wrap=True)
        log_text = '\n'.join(self.rl.messages)
        # log_message.add_row(f"{log_text}")
        
        # message = Table.grid(padding=1)
        # message.add_column()
        # message.add_column(no_wrap=True)
        # message.add_row(log_message)

        message_panel = Panel(
            Align.left(
                log_text,
                vertical="top",
            ),
            box=box.ROUNDED,
            padding=(1, 1),
            title="[b red]Log Messages",
            border_style="bright_blue",
        )
        return message_panel
    
    def update_log(self, message: str = ""):
        """Updates the log panel with a new message."""
        # self.log_message.append(message)
        self.layout["body"].update(self._make_log_panel())
    
    def _define_layout(self):
        def make_layout() -> Layout:
            """Define the layout."""
            layout = Layout(name="root")

            layout.split(
                Layout(name="header", size=3),
                Layout(name="main", ratio=1),
                Layout(name="footer", size=7),
            )
            layout["main"].split_row(
                Layout(name="side"),
                Layout(name="body", ratio=2, minimum_size=60),
            )
            layout["side"].split(
                Layout(name="box1"),
                Layout(name="box2")
            )
            return layout

        class Header:
            def __init__(self, name_layout):
                self.name_layout = name_layout
                
            def __rich__(self) -> Panel:
                grid = Table.grid(expand=True)
                grid.add_column(justify="center", ratio=1)
                grid.add_column(justify="right")
                grid.add_row(
                    f"{self.name_layout}",
                    datetime.now().ctime().replace(":", "[blink]:[/]"),
                )
                return Panel(grid, style="white on blue")

        def make_syntax() -> Syntax:
            code = """\
                print("Some msg for now")
            """
            syntax = Syntax(code, "python", line_numbers=True)
            return syntax

        self.job_progress = Progress(
            "{task.description}",
            SpinnerColumn(),
            BarColumn(),
            TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
        )
        self.job_ids = {}
        for job_name, job_length, job_id in self.init_job_list:
            self.job_ids[job_id] = self.job_progress.add_task(job_name, total=job_length)
            
            # job_progress.add_task("[green]Cooking")
            # job_progress.add_task("[magenta]Baking", total=200)
            # job_progress.add_task("[cyan]Mixing", total=400)

        total = sum(task.total if task.total is not None else 0 for task in self.job_progress.tasks)
        self.overall_progress = Progress()
        overall_task = self.overall_progress.add_task("All Jobs", total=int(total))

        progress_table = Table.grid(expand=True)
        progress_table.add_row(
            Panel(
                self.overall_progress,
                title="Overall Progress",
                border_style="green",
                padding=(2, 2),
            ),
            Panel(self.job_progress, title="[b]Jobs", border_style="red", padding=(1, 2)),
        )

        self.layout = make_layout()
        self.layout["header"].update(Header(self.name_layout))
        self.layout["body"].update(self._make_log_panel())
        self.layout["box2"].update(Panel(make_syntax(), border_style="green"))
        self.layout["box1"].update(Panel(self.layout.tree, border_style="red"))
        self.layout["footer"].update(progress_table)

    def add_job(self, job_description, job_id, total=None):
        job_rich_id = self.job_progress.add_task(job_description, total=total)
        self.job_ids[job_id] = job_rich_id
       


from datetime import datetime, timezone
from utilities_pufm.enums.enums import TimeTypes

def get_current_time(time_type: str = TimeTypes.YMD.value) -> str:
    """
    Gets the current time formatted as a string.

    This function retrieves the current date and time, formats it according to
    the provided `time_type` string, and returns it as a string. The default
    format corresponds to the `YMD` value from the `TimeTypes` enumeration.

    Args:
        time_type (str): A format string specifying how to format the date and time.
            The default is `TimeTypes.YMD.value`, typically representing a format like
            'YYYY-MM-DD'. Ensure the string conforms to Python's `strftime` directives.

    Returns:
        str: The current date and time formatted as a string according to the given format.

    Raises:
        ValueError: If the provided `time_type` is not a valid `strftime` format string.

    Example:
        >>> from utilities.time.enums import TimeTypes
        >>> get_current_time(TimeTypes.HMS.value)
        '12:34:56'
    """
    return datetime.now().strftime(time_type)

def get_time_from_seconds(seconds: int, time_type: str = TimeTypes.HMS.value) -> str:
    
    # return str(timedelta(seconds=seconds))
    return datetime.fromtimestamp(seconds, tz=timezone.utc).strftime(time_type)


# import pdfkit
# from prints.rich import custom_print
# from utilities.time.time import get_current_time

# class MigrationReport:
#     _instance = None

#     def __new__(cls):
#         # Verifica se uma instância já existe; se não, cria uma nova
#         if cls._instance is None:
#             cls._instance = super(MigrationReport, cls).__new__(cls)
#             cls._instance._initialized = False  # Inicializa apenas uma vez
#         return cls._instance
    
#     def __init__(self) -> None:
#         self.config_pdf = pdfkit.configuration(wkhtmltopdf="./report/wkhtmltopdf.exe")
#         self.report_init = []
#         self.timestamp = get_current_time()
#         tables_visits = []

#     def print_and_report(self, text: list, report: object, rich: bool = False, colors: list = ["yellow"]) -> None:
        
#         custom_print(text=''.join(text), rich=True, colors=colors)
        
#         gerar_relatorio(text=text, report=report, rich=rich)
    
# def gerar_relatorio(list_msg: list = None, local: str = None):
    
#     if local == "init":
#         report_init.extend(list_msg)
# # 
#     html_content = f"""
#     <html lang="pt-BR">
#     <head>
#         <meta charset="UTF-8">
#     </head>
#     <body>
#         <p>Prezados,</p>
#         <p>A seguir o relatório da migração em andamento realizada em {timestamp}.</p>
#         <h3>Visão Geral:</h3>
#         <p>{''.join(it+'<br/>' for it in report_init)}</p>
#         <h3>Migração das Tabelas:</h3>
#         {''.join('<h4>'+d['table']+'</h4>'+d['info']+format_list(d['coments'])+format_list(d['reports']) for d in tables_visits)}

#         <p>Atenciosamente,<br/>PROTIC,</p>
#     </body>
#     </html>
#     """

#     html_filename = os.path.join(config.config["html_dir"], f"relatorio_migration_{timestamp}.html")
#     with open(html_filename, "w", encoding="utf-8") as file:
#         file.write(html_content)
#     print(f"[REPORT]: Relatório HTML salvo como: {html_filename}")

#     pdf_filename = os.path.join(config.config["pdf_dir"], f"relatorio_migration_{timestamp}.pdf")
#     pdfkit.from_file(html_filename, pdf_filename, configuration=config_pdf)
#     print(f"[REPORT]: Relatório PDF salvo como: {pdf_filename}")


def question(question, options = ["y", "Y", "n", "N"]):
    while True:
        
        response = input(question)

        if response.strip() not in options:
            print(f"Escolha uma das opções: {options}")
        else:
            return response


import subprocess
import os

from utilities_pufm.paths.paths import windows_to_wsl_path
from pathlib import Path
from utilities_pufm.prints.rich import custom_print

def remove_ignored_files(repo_path: str, mensagem: str, windows: bool = True):
    script_path = os.path.abspath(os.path.join(Path(__file__).parent, "scripts/remove_ignored_from_git.sh"))
    
    if not os.path.isfile(script_path):
        raise FileNotFoundError(f"Script file not found: {script_path}")

    command = []
    if windows:
        command.extend(["wsl", "bash"])
        script_path = windows_to_wsl_path(script_path)
        repo_path = windows_to_wsl_path(repo_path)
    
    custom_print(f"*Repo path*: {repo_path}")
    custom_print(f"*Script path*: {script_path}")
    
    command.extend([script_path, repo_path, mensagem])
    try:
        # Executa o script com os argumentos
        subprocess.run(
            command,
            check=True,
            text=True
        )
    except subprocess.CalledProcessError as e:
        print(f"❌ Error while executing the script: {e}")



import os

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, scoped_session, Session
from typing import Dict, Tuple, Optional
from uuid import uuid4

class PgConnection:
    def __init__(self, name, pool_size=5, max_overflow=10, pool_timeout=30, pool_recycle=3600, multithread: bool = False) -> None:
        
        self.name = name
        self.session_info: Dict = {}
        self.multi_thread = multithread
        
        try:
            print(f"[POSTGRES-CONNECTION]: Iniciando processo de conexão com {name}...")

            self.connection_params = {
                "dbname": os.getenv(f"{name}_DB_NAME"),
                "user": os.getenv(f"{name}_USER"),
                "password": os.getenv(f"{name}_PASSWORD"),
                "host": os.getenv(f"{name}_HOST"),
                "port": os.getenv(f"{name}_PORT"),
            }

            if None in self.connection_params.values():
                missing = [k for k, v in self.connection_params.items() if v is None]
                raise ValueError(f"Missing environment variables: {', '.join(missing)}")

            print("[POSTGRES-CONNECTION]: Parâmetros:\n" + "\n".join([f"{key}: {val}" for key, val in self.connection_params.items()]))

            # Cria a string de conexão para SQLAlchemy
            self.connection_url = (
                f"postgresql://{self.connection_params['user']}:{self.connection_params['password']}@"
                f"{self.connection_params['host']}:{self.connection_params['port']}/{self.connection_params['dbname']}"
            )
            
            self.engine = create_engine(
                self.connection_url,
                pool_size=0,
                max_overflow=max_overflow,
                pool_timeout=pool_timeout,
                pool_recycle=pool_recycle,
                pool_pre_ping=True
            )
            
            self.Session = sessionmaker(bind=self.engine)
            self.session_factory = scoped_session(self.Session)
            print("[POSTGRES-CONNECTION]: Conexão estabelecida com sucesso usando SQLAlchemy.")

            self.poll_sessions = {
            }
            
        except Exception as e:
            print(f"[POSTGRES-CONNECTION]: Nome: {self.name}. Erro ao conectar ao Postgres: {e}")
            self.engine = None
            self.Session = None
            raise

    def get_engine(self):
        return self.engine

    def get_session(self, session_id: str) -> Tuple[str, Session]:
        if session_id in self.session_info:
            session = self.session_info[session_id]
            return session
        return None

    def get_new_session(self) -> Tuple[str, Session]:
        session_id = str(uuid4())
        if self.multi_thread:
            session = self.session_factory()
        else:
            session = self.Session()
        
        self.session_info[session_id] = [session, True]
        
        return session_id, session
    
    def close_session(self, session_id: str) -> bool:
        if session_id in self.session_info:
            dropped = self.session_info.pop(session_id, None)
            try:
                dropped[1] = False
                dropped[0].close()
                print(f"[POSTGRES-CONNECTION-{self.name}]: Session {session_id[:8]} closed.")
                return True
            except Exception as e:
                print(f"[POSTGRES-CONNECTION-error-{self.name}]: Error closing session {session_id[:8]}: {e}")
        return False

    def close_sessions(self, session_ids: list[str]) -> None:
        for session_id in session_ids:
            self.close_session(session_id)
            
    def close_all_sessions(self) -> None:
        session_ids = list(self.session_info.keys())
        self.close_sessions(session_ids)
        
    def close_all(self) -> None:
        self.close_all_sessions()
        if self.engine:
            self.engine.dispose()
            print(f"[POSTGRES-CONNECTION-{self.name}]: Engine disposed")

    def get_open_sessions_id(self) -> list[str]:
        return [session_id for session_id, (session, status) in self.session_info.items() if status]
    
    def check_for_closed_sessions(self):
        for session_id, session_status in self.session_info.items():
            if not session_status[1]:
                self.close_session(session_id)
    
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close_all()


import subprocess

from utilities_pufm.prints.rich import custom_print

class GitManager:
    _instance = None
    updated = False
    
    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(GitManager, cls).__new__(cls)
            cls._instance._initialized = False

        return cls._instance
    
    def __init__(self) -> None:
        
        if not self.updated:
            self.pull_done()
    
    def pull_done(self):
        try:
            status_output = subprocess.check_output(['git', 'status', '-uno'], text=True)
            
            if "Your branch is behind" in status_output:
                custom_print("*[YOUR BRANCH IS BEHIND*]: Executing 'git pull'", rich=True)
                subprocess.run(['git', 'pull'])
                custom_print("*[YOUR BRANCH IS UPDATED]*: Repostory updated.", rich=True)
                self.updated = True
            else:
                custom_print("Repository is *up-to-date*.", rich=True, colors=["green"])
        
        except subprocess.CalledProcessError as e:
            print(f"Erro ao executar o comando Git: {e}")
        except FileNotFoundError:
            print("Git não encontrado. Verifique se está instalado e disponível no PATH.")


import textwrap
from sqlalchemy import text
from sqlalchemy.orm import Session
import os
import subprocess
import pandas as pd
from utilities.strings.strings import join_values
from utilities.prints.rich import custom_print

def execute_sql_query(sql_query, session: Session, params: dict = None):
    try:
        if params:
            result = session.execute(text(sql_query), params)
        else:
            result = session.execute(text(sql_query))
        
        if sql_query.strip().upper().startswith("SELECT"):
            results = result.fetchall()
            column_names = result.keys()
            
            return results, column_names
        elif sql_query.strip().upper().startswith("INSERT"):
            session.commit()    
            # Importante: MySQL precisa do commit explícito
            insert_id = result.lastrowid
            
            return insert_id, True
        
        elif any(sql_query.strip().upper().startswith(cmd) for cmd in ["UPDATE", "DELETE", "DROP"]):
            session.commit()  # Commit necessário após alterações
            return result.rowcount, True
        
        elif sql_query.strip().upper().startswith("SET"):
            session.commit()
            return None, True
        
        elif sql_query.strip().upper().startswith("SHOW"):
            results = result.fetchall()
            return results, True
        
        return result.fetchall(), True
    
    except Exception as e:
        print(f"[EXECUTION] Erro na execução do script SQLAlchemy: {e}\n Query: {sql_query}")
        session.rollback()  # Faz o rollback em caso de erro
        return f"Erro: {str(e)}", None


def get_all_rows(session, tabela):
    sql_query = f"""
        SELECT *
        FROM `{tabela}`;
    """
    
    return execute_sql_query(sql_query=sql_query, session=session)

def get_column_value_by_condition(session, table, target_column, condition_column, condition_value):
    """
    Retorna os valores de uma coluna específica onde outra coluna atende a uma condição.

    Args:
        session: A sessão SQLAlchemy.
        tabela: O nome da tabela.
        target_column: A coluna cujo valor será retornado.
        condition_column: A coluna que será usada na condição.
        condition_value: O valor da condição.

    Returns:
        Uma lista de valores da target_column que atendem à condição.
    """
    
    if isinstance(condition_value, str):
        condition_value = f"'{condition_value}'"
    elif condition_value is None:
        condition_value = "NULL"
        
    sql_query = f"""
        SELECT {target_column}
        FROM `{table}`
        WHERE {condition_column} = {condition_value};
    """

    result_val, columns = execute_sql_query(sql_query=sql_query, session=session)
    
    if result_val:
        return result_val[0][0], columns  # Retorna o valor encontrado na primeira linha e primeira coluna
    else:
        return None, None

def check_value_exists(session, table_name, column_name, value):
    query_check_value = textwrap.dedent(f"""
    SELECT *
    FROM {table_name}
    WHERE {column_name} = :value
    """)

    params = {"value": value}

    results, columns = execute_sql_query(sql_query=query_check_value, session=session, params=params)
    
    return pd.DataFrame(data=results, columns=columns), True if results else False

def get_chave_primaria(session, table_name):
    query_find_primary_key = f"""
    SELECT COLUMN_NAME
    FROM information_schema.KEY_COLUMN_USAGE
    WHERE TABLE_NAME = '{table_name}'
        AND CONSTRAINT_NAME = 'PRIMARY';
    """

    results, _ = execute_sql_query(sql_query=query_find_primary_key, session=session)

    return results[0][0] if results else None

def insert_row(session, columns, list_values, tabela):
    
    # print(f"[TESTE-INSERT]: {list_values}")
    
    list_values = [
        valor.item() if isinstance(valor, pd.Series) else valor
        for valor in list_values
    ]
    
    insert_statement = textwrap.dedent(f"""
    INSERT INTO {tabela} ({', '.join(columns)}) 
    VALUES ({join_values(list_of_values=list_values)});
    """)

    # print(f"[TESTE-INSERT]: {insert_statement}")
    
    return execute_sql_query(sql_query=insert_statement, session=session)

def update_row(session, tabela, column_to_change, value_to_set, condition_column, condition_value, logic_operand):
    
    value_to_set_str = f"'{value_to_set}'" if isinstance(value_to_set, str) else value_to_set
    condition_value_str = f"'{condition_value}'" if isinstance(condition_value, str) else condition_value
    
    update_statement = textwrap.dedent(
        f"""
        UPDATE {tabela}
        SET {column_to_change} = {value_to_set_str}
        WHERE {condition_column} {logic_operand} {condition_value_str};
        """
    )
    # print(f"[TESTE]: update_statement: {update_statement}")
    return execute_sql_query(sql_query=update_statement, session=session)

########################## Obter Informações ##########################
def get_columns_names_and_types(session, table_name):
    sql = f"""
    SELECT column_name, data_type
    FROM information_schema.columns
    WHERE table_name = '{table_name}'
    ORDER BY ordinal_position;
    """
    
    resultado, _ = execute_sql_query(sql_query=sql, session=session)
    columns_name = [t[0] for t in resultado]
    columns_type = [t[1] for t in resultado]
    
    return columns_name, columns_type

########################## Backups e Restaurações ##########################
def backup_database(conn, backup_dir: str):
    
    backup_file_name = f"all_{conn.database}_backup.sql"
    backup_file_path = os.path.join(backup_dir, backup_file_name)
    
    if os.path.exists(backup_file_path):
        os.remove(backup_file_path)
        print(f"[BACKUP]: Arquivo '{os.path.abspath(backup_file_path)}' existente apagado.")

    os.environ['MYSQL_PWD'] = conn._password
    
    command = (
        f"mysqldump --skip-lock-tables --routines --add-drop-database --disable-keys "
        f"--extended-insert --comments -u {conn.user} "
        f"--host={conn.server_host} --port={conn.server_port} {conn.database} "
        f"> {backup_file_path}"
    )

    result = subprocess.run(command, shell=True, check=True)
    return backup_file_path, os.path.abspath(backup_file_path), result.returncode == 0


def backup_table(conn, tabela, backup_dir):
    backup_file_name = f"{tabela}_backup.sql"
    backup_file_path = os.path.join(backup_dir, backup_file_name)
    if os.path.exists(backup_file_path):
        os.remove(backup_file_path)
        print(
            f"[BACKUP]: Arquivo '{os.path.abspath(backup_file_path)}' existente apagado."
        )
    os.environ['MYSQL_PWD'] = conn._password
    command = (
        f"mysqldump --skip-lock-tables --routines --add-drop-table --disable-keys "
        f"--extended-insert --comments -u {conn.user} "
        f"--host={conn.server_host} --port={conn.server_port} {conn.database} {tabela} "
        f"> {backup_file_path}"
    )

    result = subprocess.run(command, shell=True, check=True)
    return backup_file_path, os.path.abspath(backup_file_path), result.returncode == 0

def restore_table(conn, backup_file):
    if not os.path.exists(backup_file):
        raise FileNotFoundError(f"O arquivo de backup '{backup_file}' não foi encontrado.")
    
    # os.environ['MYSQL_PWD'] = conn.engine._password
    url = conn.engine.url
    user = url.username
    password = url.password
    host = url.host
    port = url.port
    database = url.database
    os.environ['MYSQL_PWD'] = password
    
    command = (
        f"mysql -u {user} "
        f"--host={host} --port={port} "
        f"{database} < {backup_file}"
    )

    result = subprocess.run(command, shell=True, check=True)
    return result.returncode == 0


def restore_all_database(conn, backup_dir):
    relatorio = {}
    
    for file_name in os.listdir(backup_dir):
        if file_name.endswith("_backup.sql") and not file_name.startswith("all_"):
            backup_file = os.path.join(backup_dir, file_name)
            tabela = file_name.replace("_backup.sql", "")
            if os.path.isfile(backup_file):
                status = restore_table(conn, backup_file)
                print(f"[RESTORE]: Restaurado o backup a partir do arquivo '{backup_file}'.")
                if status:
                    relatorio[tabela] = "Sucesso!"
                    print(f"[RESTORE]: Restore do backup a partir do arquivo '{backup_file}' foi bem sucedido.")
                else:
                    relatorio[tabela] = "Falha!"
                    print(f"[RESTORE]: Restore do backup a partir do arquivo '{backup_file}' falhou.")
    
    return relatorio

def delete_all_rows(session, table_name):
    sql = f"DELETE FROM {table_name};"
    return execute_sql_query(sql_query=sql, session=session)

def drop_and_recreate_table(session, table_name):
    # Obter o script completo de criação da tabela
    create_table_script, _ = execute_sql_query(f"SHOW CREATE TABLE {table_name}", session=session)
    
    # custom_print(f"*[DROP]*: Script da *Tabela: {table_name}* criado.", rich=True, colors=["blue"])
    # Dropar a tabela
    _, status = execute_sql_query("SET foreign_key_checks = 0;", session=session)
    custom_print("*[DROP]*: Constraint desativada", rich=True, colors=["blue"])
    num_rows, _ = execute_sql_query(f"DROP TABLE IF EXISTS {table_name}", session=session)
    custom_print(f"*[DROP]*: Tabela: *{table_name}* dropada.", rich=True, colors=["blue"])
    _, status = execute_sql_query("SET foreign_key_checks = 1;", session=session)
    custom_print("*[DROP]*: Constraint ativada", rich=True, colors=["blue"])
    # print(create_table_script)
    # Recriar a tabela usando o script original
    _, recreate_status = execute_sql_query(create_table_script[0][1], session=session)
    custom_print(f"*[RECREATE]*: Tabela: *{table_name}* recriada.", rich=True, colors=["blue"])
    
    _, reset_status = execute_sql_query(f"ALTER TABLE {table_name} AUTO_INCREMENT = 1;", session=session)
    custom_print(f"*[RESET]*: AUTO_INCREMENT da tabela *{table_name}* resetado.", rich=True, colors=["blue"])
    
    return num_rows, recreate_status


import pandas as pd

def join_values(list_of_values, remove_nan: bool = False):
    
    list_of_values = [
        None
        if li is None or (remove_nan and pd.isna(li))
        else li.replace("'", "''")
        if (isinstance(li, str) and "'" in li)
        else li
        for li in list_of_values
    ]
    
    return ", ".join(
        [
            f"'{valor.strip()}'"
            if isinstance(valor, str)
            else f"{int(valor)}"
            if isinstance(valor, float) and valor.is_integer()
            else "NULL"
            if valor is None
            else f"{valor}"
            for valor in list_of_values
        ]
    )


import threading
import gc

import pandas as pd

from multiprocessing import cpu_count
from concurrent.futures import ThreadPoolExecutor
from typing import Callable
from pandas.core.groupby.generic import DataFrameGroupBy

class DataFrameThreadManager:
    def __init__(self, work_function: Callable) -> None:
        
        self.worker_function = work_function
        
    def _process_chunk(self, chunk_data: tuple) -> None:
        with threading.Lock():
            return self.worker_function(**chunk_data)
    
    def parallelize_process(self, df: pd.DataFrame | DataFrameGroupBy, n_cores=(cpu_count() - 2), params=None) -> list:
        
        # Prepara chunks
        n_rows = len(df)
        if n_rows < n_cores:
            n_cores = n_rows
            print(f"Warning: Adjusted number of threads to {n_cores} to match data size")
            
        # chunks = np.array_split(df, n_cores)
        chunk_size = n_rows // n_cores # result will be integer division
        remainder = n_rows % n_cores # remainder is the rest
        
        chunks = []
        start = 0
        if isinstance(df, pd.DataFrame):
            for i in range(n_cores):
                end = start + chunk_size + (1 if i < remainder else 0)
                if params:
                    chunks.append({
                        "t_num": i,
                        "df": df.iloc[start:end],
                        "params": params
                    })
                else:
                    chunks.append({
                        "t_num": i,
                        "df": df.iloc[start:end]
                    })
                start = end
        elif isinstance(df, DataFrameGroupBy):
            df = list(df)
            for i in range(n_cores):
                end = start + chunk_size + (1 if i < remainder else 0)  # Distribui os grupos entre os chunks
                if params:
                    chunks.append({
                        "t_num": i,
                        "df": df[start:end],
                        "params": params
                    })
                else:
                    chunks.append({
                        "t_num": i,
                        "df": df[start:end]
                    })
                start = end
        elif isinstance(df, list):
            for i in range(n_cores):
                end = start + chunk_size + (1 if i < remainder else 0)  # Distribui os grupos entre os chunks
                if params:
                    chunks.append({
                        "t_num": i,
                        "df": df[start:end],
                        "params": params
                    })
                else:
                    chunks.append({
                        "t_num": i,
                        "df": df[start:end]
                    })
                start = end
        gc.collect()
        
        with ThreadPoolExecutor(max_workers=n_cores) as executor:
            results = list(executor.map(self._process_chunk, chunks))
        
        return results